#!/usr/bin/env python
import os
import time
import multiprocessing
import logging
from pygz import PigzFile
from fbilr import utils

DEBUG = False

# CONFIDENCE_EDIT_DISTANCE = 0


class BarcodeFinder(object):
    def __init__(self):
        self.f_barcode = None
        self.f_reads = None
        self.threads = 1
        self.max_edit_distance = 5
        self.high_confident_ed = 0
        self.width = 200
        self.reads_per_batch = 10000 # 10k
        self.max_submitted_batch = self.threads * 2
        self.counter_of_summary = dict()
        self.f_matrix = None
        self.h_matrix = None
        self.f_summary = None
        self.splitted_fastq_dir = None
        self.h_fastqs = None
        self.ignore_read_name = False
    
    @staticmethod
    def extract_edge_sequences(reads, width):
        reference_sequences = []
        for read in reads:
            seq = read[1]
            length = len(seq)
            if length > width:
                seq_head = seq[:width]
                seq_tail = seq[-width:]
            else:
                seq_head = seq
                seq_tail = None
            reference_sequences.append([seq_head, seq_tail, length])
        return reference_sequences

    @staticmethod
    def _worker(references, barcodes, high_confident_ed):
        rows = []
        for seq_head, seq_tail, length in references:
            a = None  # alignment result: [loc1, loc2, ed]
            bc = None  # best barcode
            n_direction = 0  # 0: forward, 1: reverse
            n_location = 0  # 0: head 1: tail
            for bc_name, bc_seq_f, bc_seq_r in barcodes:
                tmp = utils.align(bc_seq_f, seq_head)
                if a is None or tmp[2] < a[2]:
                    a = tmp
                    bc, n_direction, n_location = bc_name, 0, 0
                    if a[2] <= high_confident_ed:
                        break
                tmp = utils.align(bc_seq_r, seq_head)
                if tmp[2] < a[2]:
                    a = tmp
                    bc, n_direction, n_location = bc_name, 1, 0
                    if a[2] <= high_confident_ed:
                        break
                if seq_tail is None:
                    continue
                tmp = utils.align(bc_seq_f, seq_tail)
                if tmp[2] < a[2]:
                    a = tmp
                    bc, n_direction, n_location = bc_name, 0, 1
                    if a[2] <= high_confident_ed:
                        break

                tmp = utils.align(bc_seq_r, seq_tail)
                if tmp[2] < a[2]:
                    a = tmp
                    bc, n_direction, n_location = bc_name, 1, 1
                    if a[2] <= high_confident_ed:
                        break
            assert a
            direction = "F" if n_direction == 0 else "R"
            start, end = a[0], a[1]
            if n_location == 1:
                offset = length - len(seq_tail)
                start, end = start + offset, end + offset
            middle_position = int(length / 2)
            if end <= middle_position:
                location = "H"
            elif start >= middle_position:
                location = "T"
            else:
                location = "M"
            row = [bc, direction, location, start, end, a[2]]
            rows.append(row)
        return rows
        
    def prepare(self):
        self.max_submitted_batch = self.threads * 2
        
        # output for metrics
        if self.f_matrix.endswith(".gz"):
            self.h_matrix = PigzFile(self.f_matrix, "wt")
        else:
            self.h_matrix = open(self.f_matrix, "w+")
            
        # load barcodes
        self.barcodes = utils.load_barcodes(self.f_barcode)
        assert len(self.barcodes) > 0
        
        # summary
        self.counter_of_summary = dict()
        for barcode in self.barcodes:
            self.counter_of_summary[barcode[0]] = 0
        self.counter_of_summary["unclassified"] = 0
        
        # output fastq
        if self.splitted_fastq_dir:
            if not os.path.exists(self.splitted_fastq_dir):
                os.mkdir(self.splitted_fastq_dir)
            self.h_fastqs = dict()
            for barcode in self.barcodes:
                bc_name = barcode[0]
                self.h_fastqs[bc_name] = open(os.path.join(self.splitted_fastq_dir, bc_name + ".fastq"), "w+")
            self.h_fastqs["unclassified"] = open(os.path.join(self.splitted_fastq_dir, "unclassified.fastq"), "w+")
                        
    def finished(self):
        if self.h_matrix:
            self.h_matrix.close()
            
        if self.f_summary:
            with open(self.f_summary, "w+") as h:
                for k, v in self.counter_of_summary.items():
                    h.write("%s\t%d\n" % (k, v))
                    
        if self.h_fastqs:
            for h in self.h_fastqs.values():
                h.close()
    
    def run_pipeline(self):
        t1 = time.time()
        self.prepare()
        self.execute()
        self.finished()
        t2 = time.time()
        t = t2 - t1
        h = int(t / 3600)
        t = t % 3600
        m = t / 60
        s = t % 60
        logging.info("spent %dh%dm%.2fs" % (h, m, s))
        
    def execute(self):
        pool = multiprocessing.Pool(self.threads, maxtasksperchild=1)
        submitted_batch_list = []
        batch_loader = utils.load_batch(self.f_reads, self.reads_per_batch)
        for batch_id, reads in enumerate(batch_loader):    
            if DEBUG and batch_id >= 16:
                break
            reference_sequences = BarcodeFinder.extract_edge_sequences(reads, self.width)
            while True:
                # Priority process finished batch
                if len(submitted_batch_list) > 0:
                    item = submitted_batch_list[0]
                    if item[2].ready():
                        assert item[2].successful()
                        self.process_results(item[1], item[2].get())
                        submitted_batch_list.pop(0)
                        logging.info("Processed batch %d" % item[0])
                # Do not submit too many tasks
                if len(submitted_batch_list) < self.max_submitted_batch:
                    args = (reference_sequences, self.barcodes, self.high_confident_ed)
                    r = pool.apply_async(BarcodeFinder._worker, args)
                    submitted_batch_list.append([batch_id, reads, r])
                    logging.info("Submitted batch %d" % batch_id)
                    break
                else:
                    time.sleep(1)
        while len(submitted_batch_list) > 0:
            item = submitted_batch_list[0]
            if item[2].ready():
                assert item[2].successful()
                self.process_results(item[1], item[2].get())
                submitted_batch_list.pop(0)
                logging.info("Processed batch %d" % item[0])
            else:
                time.sleep(1)
        pool.close()
        pool.terminate()
        pool.join()
            
    def process_results(self, reads, rows):  
        assert len(reads) == len(rows)
        for read, row in zip(reads, rows):
            name, sequence, quality = read
            length = len(sequence)
            bc, ed = row[0], row[5]
            h_fastq = None                
            if ed <= self.max_edit_distance:
                self.counter_of_summary[bc] += 1
                if self.h_fastqs:
                    h_fastq = self.h_fastqs[bc]
            else:
                self.counter_of_summary["unclassified"] += 1
                if self.h_fastqs:
                    h_fastq = self.h_fastqs["unclassified"]
            if h_fastq:
                h_fastq.write("%s\n%s\n+\n%s\n" % (name, sequence, quality))
            if self.ignore_read_name:
                row1 = [".", length] # ignore read name
            else:
                row1 = [name.split()[0][1:], length]    
            row1.extend(row)
            self.h_matrix.write("\t".join(map(str, row1)) + "\n")
                        
            
class PairEndBarcodeFinder(BarcodeFinder):
    @staticmethod
    def _worker(references, barcodes, high_confident_ed):
        rows = []
        for seq_head, seq_tail, length in references:
            if length < len(seq_head) * 2:
                rows.append([".", ".", ".", -1, -1, -1, ".", ".", ".", -1, -1, -1])
                continue
            a_head = None # alignment result: [start, end, ed]
            a_tail = None
            bc_head = None # best barcode
            bc_tail = None
            n_direction_head = 0 # 0: forward, 1: reverse
            n_direction_tail = 0
            n_location_head = 0 # 0: head 1: tail
            n_location_tail = 0
            for bc_name, bc_seq_f, bc_seq_r in barcodes:
                tmp = utils.align(bc_seq_f, seq_head)
                if a_head is None or tmp[2] < a_head[2]:
                    a_head = tmp
                    bc_head, n_direction_head, n_location_head = bc_name, 0, 0
                    if a_head[2] <= high_confident_ed:
                        break
                tmp = utils.align(bc_seq_r, seq_head)
                if tmp[2] < a_head[2]:
                    a_head = tmp
                    bc_head, n_direction_head, n_location_head = bc_name, 1, 0
                    if a_head[2] <= high_confident_ed:
                        break
            for bc_name, bc_seq_f, bc_seq_r in barcodes:
                tmp = utils.align(bc_seq_f, seq_tail)
                if a_tail is None or tmp[2] < a_tail[2]:
                    a_tail = tmp
                    bc_tail, n_direction_tail, n_location_tail = bc_name, 0, 1
                    if a_tail[2] <= high_confident_ed:
                        break
                tmp = utils.align(bc_seq_r, seq_tail)
                if tmp[2] < a_tail[2]:
                    a_tail = tmp
                    bc_tail, n_direction_tail, n_location_tail = bc_name, 1, 1
                    if a_tail[2] <= high_confident_ed:
                        break
            assert a_head
            assert a_tail
            direction_head = "F" if n_direction_head == 0 else "R"
            direction_tail = "F" if n_direction_tail == 0 else "R"
            start_head, end_head = a_head[0], a_head[1]
            start_tail, end_tail = a_tail[0], a_tail[1]
            assert n_location_head == 0
            assert n_location_tail == 1
            offset = length - len(seq_tail)
            start_tail, end_tail = start_tail + offset, end_tail + offset
            row = [bc_head, direction_head, "H", start_head, end_head, a_head[2], 
                   bc_tail, direction_tail, "T", start_tail, end_tail, a_tail[2]]
            rows.append(row)
        return rows
    
    def prepare(self):
        self.max_submitted_batch = self.threads * 2
        # output for metrics
        if self.f_matrix.endswith(".gz"):
            self.h_matrix = PigzFile(self.f_matrix, "wt")
        else:
            self.h_matrix = open(self.f_matrix, "w+")
        # load barcodes
        self.barcodes = utils.load_barcodes(self.f_barcode)
        assert len(self.barcodes) > 0
        assert self.splitted_fastq_dir is None # pair-end mode do not support output fastq files
        assert self.f_summary is None # pair-end mode do not support summary
    
    def execute(self):
        pool = multiprocessing.Pool(self.threads, maxtasksperchild=1)
        submitted_batch_list = []
        batch_loader = utils.load_batch(self.f_reads, self.reads_per_batch)
        for batch_id, reads in enumerate(batch_loader):    
            if DEBUG and batch_id >= 16:
                break
            reference_sequences = BarcodeFinder.extract_edge_sequences(reads, self.width)
            while True:
                # Priority process finished batch
                if len(submitted_batch_list) > 0:
                    item = submitted_batch_list[0]
                    if item[2].ready():
                        assert item[2].successful()
                        self.process_results(item[1], item[2].get())
                        submitted_batch_list.pop(0)
                        logging.info("Processed batch %d" % item[0])
                # Do not submit too many tasks
                if len(submitted_batch_list) < self.max_submitted_batch:
                    args = (reference_sequences, self.barcodes, self.high_confident_ed)
                    r = pool.apply_async(PairEndBarcodeFinder._worker, args)
                    submitted_batch_list.append([batch_id, reads, r])
                    logging.info("Submitted batch %d" % batch_id)
                    break
                else:
                    time.sleep(1)
        while len(submitted_batch_list) > 0:
            item = submitted_batch_list[0]
            if item[2].ready():
                assert item[2].successful()
                self.process_results(item[1], item[2].get())
                submitted_batch_list.pop(0)
                logging.info("Processed batch %d" % item[0])
            else:
                time.sleep(1)
        pool.close()
        pool.terminate()
        pool.join()
               
    def process_results(self, reads, rows):  
        assert len(reads) == len(rows)
        for read, row in zip(reads, rows):
            name, sequence, quality = read
            length = len(sequence)
            if self.ignore_read_name:
                row1 = [".", length] # ignore read name
            else:
                row1 = [name.split()[0][1:], length]
            row1.extend(row)
            self.h_matrix.write("\t".join(map(str, row1)) + "\n")



#!/usr/bin/env python

from pygz import PigzFile


class Hit(object):
    def __init__(self, name, direction, location, start, end, ed):
        self.name = name
        self.direction = direction
        self.location = location
        self.start = start
        self.end = end
        self.ed = ed
        
    def __str__(self) -> str:
        return "%s\t%s\t%s\t%d\t%d\t%d" % (self.name, self.direction, self.location, 
                                           self.start, self.end, self.ed)
        

class MatrixRecord(object):
    def __init__(self, name, length, hits, read=None):
        self.name = name
        self.length = length
        self.hits = hits
        self.read = read
    
    def __str__(self):
        items = [self.name, self.length]
        for hit in self.hits:
            items.append(hit.name)
            items.append(hit.direction)
            items.append(hit.location)
            items.append(hit.start)
            items.append(hit.end)
            items.append(hit.ed)
        # if self.read is not None:
        #     pass
        return "\t".join(map(str, items))


class MatrixReader(object):
    def open(path):
        with open(path) as f:
            for line in f:
                
                row = line.strip("\n").split("\t")
                
                name, length = row[0], int(row[1])
                
                hits = []
                n, m = int((len(row) - 2) / 6), (len(row) - 2) % 6
                assert n > 0
                assert m == 0 or m == 4
                for i in range(n):
                    bc, direction, loc, start, end, ed = row[i * 6 + 2:(i + 1) * 6 + 2]
                    hit = Hit(name=bc, 
                              direction=direction, 
                              location=loc, 
                              start=int(start), 
                              end=int(end), 
                              ed=int(ed))
                    hits.append(hit)
                    
                read = None
                if m == 4:
                    pass
                    
                record = MatrixRecord(name=name, length=length, hits=hits, read=read)
                    
                yield record
        

class Matrix2Record(object):
    def __init__(self):
        # read
        self.read_name = None
        self.read_length = None
        # head
        self.head_barcode = "."
        self.head_direction = "."
        self.head_location = "."
        self.head_start = -1
        self.head_end = -1
        self.head_ed = -1
        # tail
        self.tail_barcode = "."
        self.tail_direction = "."
        self.tail_location = "."
        self.tail_start = -1
        self.tail_end = -1
        self.tail_ed = -1
        
    def __str__(self):
        row = [self.read_name, self.read_length, 
               self.head_barcode, self.head_direction, self.head_location, 
               self.head_start, self.head_end, self.head_ed, 
               self.tail_barcode, self.tail_direction, self.tail_location, 
               self.tail_start, self.tail_end, self.tail_ed]
        s = "\t".join(map(str, row))
        return s


class Matrix2Reader(object):
    @classmethod
    def open(cls, path):
        with PigzFile(path, "rt") as f:
            for line in f:
                row = line.strip("\n").split("\t")
                record = Matrix2Record()
                record.read_name = row[0]
                record.read_length = int(row[1])
                if row[2] != ".":
                    record.head_barcode = row[2]
                    record.head_direction = row[3]
                    record.head_location = row[4]
                    record.head_start = int(row[5])
                    record.head_end = int(row[6])
                    record.head_ed = int(row[7])
                if row[8] != ".":
                    record.tail_barcode = row[8]
                    record.tail_direction = row[9]
                    record.tail_location = row[10]
                    record.tail_start = int(row[11])
                    record.tail_end = int(row[12])
                    record.tail_ed = int(row[13])
                yield record



import edlib


def find_all(que, ref, offset=0, max_ed=5, min_length=10):
    results = []
    if len(ref) < min_length:
        return results
    r = edlib.align(que, ref, task="locations", mode="HW")
    ed = r["editDistance"]
    if ed <= max_ed:
        locs = []
        for x, y in r["locations"]:
            y += 1
            if len(locs) == 0:
                locs.append([x, y])
            else:
                if x >= locs[-1][1]:
                    locs.append([x, y])
        for x, y in locs:
            results.append([offset + x, offset + y, ed])
            
        y0 = 0
        data = []
        for x, y in locs:
            if x - y0 >= min_length:
                data.append([ref[y0:x], offset + y0])
            y0 = y
        if len(ref) - y0 >= min_length:
            data.append([ref[y0:len(ref)], offset + y0])
        for item in data:
            for x, y, ed in find_all(que, item[0], item[1], max_ed, min_length):
                results.append([x, y, ed])
    return results


#!/usr/bin/env python
import optparse
import logging
from fbilr.finders import BarcodeFinder
from fbilr.finders import PairEndBarcodeFinder



def main():
    
    logging.basicConfig(level=logging.INFO, format = '%(asctime)s - %(levelname)s: %(message)s')

    
    # Parameters
    
    parser = optparse.OptionParser(usage="find_barcode.py [options] barcodes.fa reads.fq.gz")
    
    parser.add_option("-t", "--threads", dest="threads", type="int", default=1, 
                    help="Number of threads used. [%default]")
    parser.add_option("-e", "--max-edit-istance", dest="max_edit_distance", default=5,
                    help="Max edit distance of pass barcode. [%default]")
    parser.add_option("-c", "--confident-edit-distance", dest="confident_ed", default=0, 
                      help="Edit distance threshold for confident barcode. When edit distance <= threshold, stop to find other barcodes. [%default]")
    parser.add_option("-w", "--width", dest="width", default=200,
                    help="Find barcode sequences within WIDTH of the boundary. [%default]")
    parser.add_option("-m", "--matrix", dest="matrix", default=None,
                    help="Path of metrics file. [%default]")
    parser.add_option("-i", "--ignore-read-name", dest="ignore_read_name", action="store_true", default=False,
                      help="Ignore read name to save storage. [%default]")
    
    group = optparse.OptionGroup(parser, "For single-end barcode")
    group.add_option("-s", "--summary", dest="summary", default=None,
                    help="Path of stats file. [%default]")
    group.add_option("-d", "--outdir", dest="outdir",
                    help="Output directory for splitted fastq files. [%default]")
    parser.add_option_group(group)
    
    group = optparse.OptionGroup(parser, "For paired-end barcode")
    group.add_option("-p", "--pair-end", dest="is_pair_end", action="store_true", default=False,
                    help="Find barcode at head and tail seperately. [%default]")
    parser.add_option_group(group)
    
    options, args = parser.parse_args()
    
    f_barcode = args[0]
    f_reads = args[1]
    f_matrix = options.matrix
    f_summary = options.summary
    outdir = options.outdir
    threads = options.threads
    max_edit_distance = options.max_edit_distance
    high_confident_ed = options.confident_ed
    width = options.width
    is_pair_end = options.is_pair_end
    ignore_read_name = options.ignore_read_name
    
    # Run
    
    if is_pair_end:
        obj = PairEndBarcodeFinder()
    else:
        obj = BarcodeFinder()
    obj.f_barcode = f_barcode
    obj.f_reads = f_reads
    obj.threads = threads
    obj.f_matrix = f_matrix
    obj.f_summary = f_summary
    obj.splitted_fastq_dir = outdir
    obj.max_edit_distance = max_edit_distance
    obj.high_confident_ed = high_confident_ed
    obj.width = width
    obj.ignore_read_name = ignore_read_name
    
    obj.run_pipeline()
    

if __name__ == "__main__":
    main()


#!/usr/bin/env python
import sys
import gzip
from Bio import SeqIO
import edlib
import pysam
from pygz import PigzFile


def load_barcodes(path):
    barcodes = []
    if path.endswith(".gz"):
        f = gzip.open(path, "rt")
    else:
        f = open(path)
    for record in SeqIO.parse(f, "fasta"):
        bc_name = record.name
        bc_seq_f = str(record.seq)
        bc_seq_r = str(record.seq.reverse_complement())
        barcodes.append([bc_name, bc_seq_f, bc_seq_r])
    f.close()
    return barcodes
    
    
def load_reads(path):
    count = 0 # loaded read number
    if path == "-":
        f = sys.stdin
    if path.endswith(".gz"):
        f = PigzFile(path, "rt")
    else:
        f = open(path)
    name = None
    seq = None
    qua = None
    for i, line in enumerate(f):
        j = i % 4
        if j == 0:
            name = line[:-1]
        elif j == 1:
            seq = line[:-1]
        elif j == 3:
            qua = line[:-1]
            yield [name, seq, qua]
            count += 1
    f.close()
    
    
def load_batch(path, reads_per_batch=100):
    reads = list()
    with pysam.FastqFile(path) as f:
        for read in f:
            reads.append([read.name, read.sequence, read.quality])
            if len(reads) >= reads_per_batch:
                yield reads
                reads = list()
    if len(reads) > 0:
        yield reads
    
    # reads = None
    # for read in load_reads(path):
    #     if reads is None:
    #         reads = [read]
    #     else:
    #         reads.append(read)
    #     if len(reads) >= reads_per_batch:
    #         yield reads
    #         reads = None
    # if reads is not None:
    #     yield reads


def cut_edge_sequence(seq, width):
    head, tail = "", ""
    if len(seq) > 0 and width > 0:
        width = min(width, len(seq))
        head, tail = seq[:width], seq[-width:]
    return [len(seq), head, tail]

        
def align(query, reference):
    # assert len(query) <= len(reference)
    a = edlib.align(query, reference, task="locations", mode="HW")
    ed = a["editDistance"]
    start, end = a["locations"][0]
    end += 1
    return start, end, ed


#!/usr/bin/env python
import glob
import setuptools

scripts = []
for s in glob.glob("src/scripts/*"):
    scripts.append(s)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fbilr",
    version="1.1.0",
    author="Zonggui Chen",
    author_email="ggchenzonggui@qq.com",
    description="Find barcode in long reads",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ckenen/fbilr",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix"
    ],
    scripts=scripts,
    install_requires=['edlib', 'biopython', 'numpy', 'matplotlib', 'pygz'],
    test_suite="tests",
    python_requires=">=3.6",
)



#!/usr/bin/env python
import sys
import multiprocessing as mp
import time
import random


def worker(i):
    print(i)
    time.sleep(random.random())
    return i

def main():
    fw = open("a.txt", "w+")
    threads = 2
    max_submitted_task = threads * 2
    pool = mp.Pool(threads, maxtasksperchild=1)
    submitted_task_list = []
    for i in range(30):
        while True:
            if len(submitted_task_list) > 0:
                r = submitted_task_list[0]
                if r.ready():
                    assert r.successful()
                    fw.write("%s\n" % r.get())
                    submitted_task_list.pop(0)
                
            if len(submitted_task_list) < max_submitted_task:
                r = pool.apply_async(worker, (i, ))
                submitted_task_list.append(r)
                break
            else:
                time.sleep(1)

    while len(submitted_task_list) > 0:
        r = submitted_task_list[0]
        if r.ready():
            assert r.successful()
            fw.write("%s\n" % r.get())
            submitted_task_list.pop(0)
        else:
            time.sleep(1)

    pool.close()
    pool.join()

    fw.close()

main()




#!/usr/bin/env python
import sys
from collections import defaultdict
import numpy as np
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from pygz import GzipFile


MAX_LENGTH = 5000
BIN_WIDTH = 100
BIN_COUNT = int(MAX_LENGTH / BIN_WIDTH)
MAX_EDIT_DISTANCE = 5


def plot_edit_distance_distribution(rows, outfile):
    counter = defaultdict(int)
    for row in rows:
        counter[row[6]] += 1
    ys = np.zeros(15)
    for i in range(len(ys)):
        ys[i] = counter[i]
    xs = np.arange(len(ys))
    ys = ys * 100 / len(rows)

    plt.figure(figsize=(4, 3))
    plt.bar(xs, ys, color="C0", lw=1, edgecolor="black")
    plt.xticks(xs)
    plt.xlim(min(xs) - 0.5, max(xs) + 0.5)
    plt.ylim(0, 100)
    plt.xlabel("Edit distance")
    plt.ylabel("Percentage (%)")
    plt.tight_layout()
    plt.savefig(outfile, dpi=300)
    plt.close()


def plot_orient_loc_distribution(rows, outfile):
    counter = dict()
    for row in rows:
        d = row[6]
        if d not in counter:
            counter[d] = defaultdict(int)
        counter[d][(row[2], row[3])] += 1

    fig, axs = plt.subplots(4, 5, figsize=(12, 10), sharex=True, sharey=True)

    for i in range(20):
        m = int(i / 5)
        n = i % 5
        if i in counter:
            d = counter[i]
        else:
            d = defaultdict(int)
        keys = [("F", "H"), ("F", "T"), ("R", "H"), ("R", "T"), ]
        xticks = ["FH", "FT", "RH", "RT"]
        counts = []
        for key in keys:
            count = d[key]
            counts.append(count)
        ys = np.array(counts) * 100 / sum(counts)
        xs = np.arange(len(ys))
        plt.sca(axs[m][n])
        plt.title("ED=%d, N=%d" % (i, sum(counts)))
        plt.bar(xs, ys, color="C0", edgecolor="black", lw=1)
        plt.ylim(0, 100)
        plt.xticks(xs, xticks)
        plt.ylabel("Percentage (%)")

    plt.tight_layout()
    plt.savefig(outfile, dpi=300)
    plt.close()


def plot_length_distribution(rows, outfile):
    array = []
    for i in range(MAX_EDIT_DISTANCE):
        array.append(np.zeros(BIN_COUNT))
    counts = np.zeros(BIN_COUNT)

    for row in rows:
        i = int(row[0] / BIN_WIDTH)
        if 0 <= i < len(counts):
            counts[i] += 1
            if row[6] < MAX_EDIT_DISTANCE:
                array[row[6]][i] += 1

    counts = counts / BIN_WIDTH
    for i in range(len(array)):
        array[i] = array[i] / BIN_WIDTH

    xs = (np.arange(BIN_COUNT) + 0.5) * BIN_WIDTH

    fig, axs = plt.subplots(2, 1, figsize=(6, 6), sharex=True)

    plt.sca(axs[0])
    plt.plot(xs, counts, label="All")
    plt.plot(xs, sum(array), label="ED<=%d" % MAX_EDIT_DISTANCE)
    plt.ylabel("Read numbers")
    plt.legend()
    
    ratios = sum(array) / counts
    plt.sca(axs[1])
    plt.plot(xs, ratios)
    plt.ylim(0, 1)
    plt.ylabel("Ratio of barcode/total")
    plt.xlabel("Read length (nt)")

    plt.tight_layout()
    plt.savefig(outfile, dpi=300)
    plt.close()


def plot_length_edit_distance(rows, outfile):
    counter = dict()
    for row in rows:
        i = int(row[0] / BIN_WIDTH)
        if i not in counter:
            counter[i] = defaultdict(int)
        counter[i][row[6]] += 1
    xs = (np.arange(BIN_COUNT) + 0.5) * BIN_WIDTH

    values = np.zeros(BIN_COUNT)
    for i in range(BIN_COUNT):
        if i in counter:
            v = sum([x for x in counter[i].values()])
        else:
            v = 0
        values[i] = v

    plt.figure(figsize=(6, 4))
    bottoms = np.zeros(BIN_COUNT)
    for ed in range(10):
        ys = np.zeros(BIN_COUNT)
        for i in range(BIN_COUNT):
            if i in counter:
                y = counter[i][ed]
            else:
                y = 0
            ys[i] = y
        ys = ys / values
        plt.bar(xs, ys, width=BIN_WIDTH, bottom=bottoms, label=str(ed))
        bottoms += ys
    plt.bar(xs, 1 - bottoms, width=BIN_WIDTH,
            bottom=bottoms, color="lightgrey", label=">9")
    plt.xlabel("Read length (nt)")
    plt.ylabel("Proportion")
    plt.xlim(0, MAX_LENGTH)
    plt.ylim(0, 1)
    plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
    plt.tight_layout()
    plt.savefig(outfile, dpi=300)
    plt.close()
    

def plot_location(rows, outfile):
    counter_f_h = defaultdict(int)
    counter_f_t = defaultdict(int)
    counter_r_h = defaultdict(int)
    counter_r_t = defaultdict(int)

    for row in rows:
        if row[6] > MAX_EDIT_DISTANCE: # edit distance
            continue
        orient = row[2]
        length = row[0]
        start = row[4]
        end = row[5]
        dis1 = start
        dis2 = length - end
        if orient == "F":
            counter_f_h[dis1] += 1
            counter_f_t[dis2] += 1
        else:
            counter_r_h[dis1] += 1
            counter_r_t[dis2] += 1

    fig, axs = plt.subplots(2, 2, figsize=(16, 6))

    plt.sca(axs[0][0])
    plt.title("Barcodes in forward")
    xs = np.arange(0, 151)
    ys = [counter_f_h[x] for x in xs]
    t = sum(counter_f_h.values())
    ys = np.array(ys) * 100 / t
    plt.bar(xs, ys, width=1)
    plt.xlim(min(xs) - 1, max(xs) + 1)
    plt.ylabel("Percentage (%)")
    plt.xlabel("Distance to head end")
    plt.tight_layout()

    plt.sca(axs[0][1])
    plt.title("Barcodes in forward")
    xs = np.arange(-150, 1)
    ys = [counter_f_t[-x] for x in xs]
    t = sum(counter_f_t.values())
    ys = np.array(ys) * 100 / t
    plt.bar(xs, ys, width=1)
    plt.xlim(min(xs) - 1, max(xs) + 1)
    plt.xlabel("Distance to tail end")
    plt.tight_layout()

    plt.sca(axs[1][0])
    plt.title("Barcodes in reverse")
    xs = np.arange(0, 151)
    ys = [counter_r_h[x] for x in xs]
    t = sum(counter_r_h.values())
    ys = np.array(ys) * 100 / t
    plt.bar(xs, ys, width=1)
    plt.xlim(min(xs) - 1, max(xs) + 1)
    plt.ylabel("Percentage (%)")
    plt.xlabel("Distance to head end")
    plt.tight_layout()

    plt.sca(axs[1][1])
    plt.title("Barcodes in reverse")
    xs = np.arange(-150, 1)
    ys = [counter_r_t[-x] for x in xs]
    t = sum(counter_r_t.values())
    ys = np.array(ys) * 100 / t
    plt.bar(xs, ys, width=1)
    plt.xlim(min(xs) - 1, max(xs) + 1)
    plt.xlabel("Distance to tail end")
    plt.tight_layout()

    plt.savefig(outfile, dpi=300)


def main():
    infile, prefix = sys.argv[1:]

    rows = []
    if infile.endswith(".gz"):
        f = GzipFile(infile, "rt")
    else:
        f = open(infile)
    for line in f:
        row = line.strip("\n").split("\t")
        row[1] = int(row[1])
        row[5] = int(row[5])
        row[6] = int(row[6])
        row[7] = int(row[7])
        rows.append(row[1:]) # remove read name
    f.close()
            
    plot_edit_distance_distribution(rows, prefix + ".ed_dis.pdf")
    plot_orient_loc_distribution(rows, prefix + ".orient_loc.pdf")
    plot_length_distribution(rows, prefix + ".len.pdf")
    plot_length_edit_distance(rows, prefix + ".len_ed.pdf")
    plot_location(rows, prefix + ".loc_dis.pdf")


if __name__ == '__main__':
    main()
